import cocotb
from cocotb.clock import Clock
from cocotb.triggers import Timer
from cocotb.triggers import RisingEdge
from cocotbext.axi import AxiLiteBus, AxiLiteMaster, AxiLiteRam, AxiStreamBus, AxiStreamSink

import logging

from decimal import Decimal

CLK_PERIOD_NS = 10

class TB:
    def __init__(self, dut):
        self.dut = dut

        self.log = logging.getLogger("cocotb.tb")
        self.log.setLevel(logging.DEBUG)

        cocotb.start_soon(Clock(dut.clk, CLK_PERIOD_NS, units="ns").start())

        self.axil_master = AxiLiteMaster(AxiLiteBus.from_prefix(dut, "s_regs_axil"), dut.clk, dut.rst)
        self.axis_sink = AxiStreamSink(AxiStreamBus.from_prefix(dut, "m_stream_axis"), dut.clk, dut.rst)
        self.axil_ram = AxiLiteRam(AxiLiteBus.from_prefix(dut, "m_dma_axil"), dut.clk, dut.rst, size=2**16)

    async def cycle_reset(self):
        self.dut.rst.setimmediatevalue(0)
        await RisingEdge(self.dut.clk)  # type: ignore
        await RisingEdge(self.dut.clk)  # type: ignore
        self.dut.rst.value = 1
        await RisingEdge(self.dut.clk)  # type: ignore
        await RisingEdge(self.dut.clk)  # type: ignore
        self.dut.rst.value = 0
        await RisingEdge(self.dut.clk)  # type: ignore
        await RisingEdge(self.dut.clk)  # type: ignore

@cocotb.test()
async def test_simple(dut):
    tb = TB(dut)

    await tb.cycle_reset()

    tb.axil_ram.write_dword(0x00000000, 0x00001000)
    tb.axil_ram.write_dword(0x00000004, 64)
    tb.axil_ram.write_dword(0x00000008, 0)
    tb.axil_ram.write_dword(0x0000000c, 0x00000010)

    tb.axil_ram.write_dword(0x00000010, 0x00002000)
    tb.axil_ram.write_dword(0x00000014, 64)
    tb.axil_ram.write_dword(0x00000018, 0)
    tb.axil_ram.write_dword(0x0000001c, 0x00000000)

    test_data = bytearray([x % 256 for x in range(256)])

    tb.axil_ram.write(0x1000, test_data)
    tb.axil_ram.write(0x2000, test_data)


    # Bring out of sw reset
    await tb.axil_master.write_dword(0x0c, 0x00)

    # Write start address
    await tb.axil_master.write_dword(0x0, 0x00000000)

    # Write last address
    await tb.axil_master.write_dword(0x4, 0x00000010)

    data1 = await tb.axis_sink.recv()
    data2 = await tb.axis_sink.recv()

    assert data1.tdata == test_data[0:64]
    assert data2.tdata == test_data[0:64]


    await Timer(Decimal(CLK_PERIOD_NS * 400), units='ns')


@cocotb.test()
async def test_desc_count(dut):
    tb = TB(dut)

    await tb.cycle_reset()

    tb.axil_ram.write_dword(0x00000000, 0x00001000)
    tb.axil_ram.write_dword(0x00000004, 64)
    tb.axil_ram.write_dword(0x00000008, 0)
    tb.axil_ram.write_dword(0x0000000c, 0x00000010)

    tb.axil_ram.write_dword(0x00000010, 0x00002000)
    tb.axil_ram.write_dword(0x00000014, 64)
    tb.axil_ram.write_dword(0x00000018, 0)
    tb.axil_ram.write_dword(0x0000001c, 0x00000000)

    test_data = bytearray([x % 256 for x in range(256)])

    tb.axil_ram.write(0x1000, test_data)
    tb.axil_ram.write(0x2000, test_data)


    # Bring out of sw reset
    await tb.axil_master.write_dword(0x0c, 0x00)

    # Write start address
    await tb.axil_master.write_dword(0x0, 0x00000000)

    # Write last address
    await tb.axil_master.write_dword(0x4, 0x00000010)

    await tb.axis_sink.recv()
    await tb.axis_sink.recv()

    count = await tb.axil_master.read_dword(0x8)

    assert count == 2


@cocotb.test()
async def test_sw_reset(dut):
    tb = TB(dut)

    await tb.cycle_reset()

    tb.axil_ram.write_dword(0x00000000, 0x00001000)
    tb.axil_ram.write_dword(0x00000004, 64)
    tb.axil_ram.write_dword(0x00000008, 0)
    tb.axil_ram.write_dword(0x0000000c, 0x00000010)

    tb.axil_ram.write_dword(0x00000010, 0x00002000)
    tb.axil_ram.write_dword(0x00000014, 64)
    tb.axil_ram.write_dword(0x00000018, 0)
    tb.axil_ram.write_dword(0x0000001c, 0x00000000)

    test_data = bytearray([x % 256 for x in range(256)])

    tb.axil_ram.write(0x1000, test_data)
    tb.axil_ram.write(0x2000, test_data)


    # Bring out of sw reset
    await tb.axil_master.write_dword(0x0c, 0x00)

    # Write start address
    await tb.axil_master.write_dword(0x0, 0x00000000)

    # Write last address
    await tb.axil_master.write_dword(0x4, 0x00000010)

    await tb.axis_sink.recv()
    await tb.axis_sink.recv()

    # Read registers
    assert await tb.axil_master.read_dword(0x0) == 0x10
    assert await tb.axil_master.read_dword(0x4) == 0x10
    assert await tb.axil_master.read_dword(0x8) == 2

    # Reset
    await tb.axil_master.write_dword(0xc, 1)
    await tb.axil_master.write_dword(0xc, 0)

    # Registers should all be 0
    assert await tb.axil_master.read_dword(0x0) == 0
    assert await tb.axil_master.read_dword(0x4) == 0
    assert await tb.axil_master.read_dword(0x8) == 0

