module m2s_dma #(
    parameter AXIL_DATA_WIDTH=32,
    parameter AXIL_ADDR_WIDTH=32,
    parameter AXIS_DATA_WIDTH=8,
    localparam AXIL_STRB_WIDTH=(AXIL_DATA_WIDTH+7)/8,
    localparam AXIS_KEEP_WIDTH=(AXIS_DATA_WIDTH+7)/8
)(
    input i_clk,
    input i_rst,

    input  wire                         s_cpuif_req,
    input  wire                         s_cpuif_req_is_wr,
    input  wire [3:0]                   s_cpuif_addr,
    input  wire [31:0]                  s_cpuif_wr_data,
    input  wire [31:0]                  s_cpuif_wr_biten,
    output wire                         s_cpuif_req_stall_wr,
    output wire                         s_cpuif_req_stall_rd,
    output wire                         s_cpuif_rd_ack,
    output wire                         s_cpuif_rd_err,
    output wire [31:0]                  s_cpuif_rd_data,
    output wire                         s_cpuif_wr_ack,
    output wire                         s_cpuif_wr_err,

    axil_intf.MASTER                    m_axil,
    axis_intf.MASTER                    m_axis
);

m2s_dma_regs_pkg::m2s_dma_regs__in_t hwif_in;
m2s_dma_regs_pkg::m2s_dma_regs__out_t hwif_out;

m2s_dma_regs u_m2s_dma_regs (
    .clk                    (i_clk),
    .rst                    (i_rst),

    .s_cpuif_req            (s_cpuif_req),
    .s_cpuif_req_is_wr      (s_cpuif_req_is_wr),
    .s_cpuif_addr           (s_cpuif_addr),
    .s_cpuif_wr_data        (s_cpuif_wr_data),
    .s_cpuif_wr_biten       (s_cpuif_wr_biten),
    .s_cpuif_req_stall_wr   (s_cpuif_req_stall_wr),
    .s_cpuif_req_stall_rd   (s_cpuif_req_stall_rd),
    .s_cpuif_rd_ack         (s_cpuif_rd_ack),
    .s_cpuif_rd_err         (s_cpuif_rd_err),
    .s_cpuif_rd_data        (s_cpuif_rd_data),
    .s_cpuif_wr_ack         (s_cpuif_wr_ack),
    .s_cpuif_wr_err         (s_cpuif_wr_err),

    .hwif_in                (hwif_in),
    .hwif_out               (hwif_out)
);


logic sw_reset;
assign sw_reset = hwif_out.ctrl.swreset.value;

logic saw_arready, saw_arready_next;
logic buffer_empty, buffer_empty_next;
logic buffer_full, buffer_full_next;

logic [31:0] buffer, buffer_next;

logic [31:0] current_desc_addr, current_desc_addr_next;

assign hwif_in.end_addr.d.hwclr = sw_reset;


logic [31:0] desc_base_addr, desc_base_addr_next;
logic [31:0] desc_length, desc_length_next;
logic [31:0] desc_next_addr, desc_next_addr_next;


logic [31:0] state_counter, state_counter_next;

logic [31:0] desc_count, desc_count_next;
logic [31:0] width_convert_count, width_convert_count_next;

enum logic [2:0] {
    IDLE,
    DESC_FETCH,
    READ,
    STREAM,
    CONTINUE
} state, state_next;

always_ff @(posedge i_clk) begin
    if (i_rst || sw_reset) begin
        current_desc_addr <= '0;
        desc_base_addr <= '0;
        desc_next_addr <= '0;
        desc_length <= '0;
        desc_count <= '0;
        state_counter <= '0;
        state <= IDLE;

        saw_arready <= '0;
        buffer_empty <= '0;
        buffer_full <= '0;
        buffer <= '0;
        width_convert_count <= '0;
    end else begin
        current_desc_addr <= current_desc_addr_next;
        desc_base_addr <= desc_base_addr_next;
        desc_next_addr <= desc_next_addr_next;
        desc_length <= desc_length_next;
        desc_count <= desc_count_next;
        state_counter <= state_counter_next;
        state <= state_next;

        saw_arready <= saw_arready_next;
        buffer_empty <= buffer_empty_next;
        buffer_full <= buffer_full_next;
        buffer <= buffer_next;
        width_convert_count <= width_convert_count_next;
    end
end

always_comb begin
    desc_base_addr_next = desc_base_addr;
    desc_next_addr_next = desc_next_addr;
    desc_length_next = desc_length;
    desc_count_next = desc_count;
    current_desc_addr_next = current_desc_addr;
    state_counter_next = state_counter;
    state_next = state;

    width_convert_count_next = width_convert_count;
    buffer_empty_next = buffer_empty;
    buffer_full_next = buffer_full;
    buffer_next = buffer;
    hwif_in.dma_done.d.next = desc_count;

    if (sw_reset) begin
        hwif_in.curr_addr.d.next = '0;
    end else begin
        hwif_in.curr_addr.d.next = hwif_out.curr_addr.d.value;
    end

    m_axis.tvalid = '0;
    m_axis.tdata = '0;
    m_axis.tkeep = '0;
    m_axis.tlast = '0;
    m_axis.tid = '0;
    m_axis.tdest = '0;
    m_axis.tuser = '0;

    m_axil.awvalid  = '0;
    m_axil.awaddr   = '0;
    m_axil.awprot   = '0;
    m_axil.wvalid   = '0;
    m_axil.wdata    = '0;
    m_axil.wstrb    = '0;
    m_axil.bready   = '0;
    m_axil.arvalid  = '0;
    m_axil.araddr   = '0;
    m_axil.arprot   = '0;
    m_axil.rready   = '0;

    saw_arready_next = saw_arready;

    case (state)
        IDLE: begin
            if (hwif_out.end_addr.d.swmod) begin
                state_next = DESC_FETCH;
                current_desc_addr_next = hwif_out.curr_addr.d.value;
                hwif_in.curr_addr.d.next = '0;
            end
        end

        DESC_FETCH: begin
            m_axil.araddr = current_desc_addr + (state_counter << 2);

            m_axil.arvalid = saw_arready ? '0 : '1;
            saw_arready_next = m_axil.arready ? '1 : saw_arready;

            m_axil.rready = '1;

            if (m_axil.rvalid) begin
                state_counter_next = state_counter + 1;
                saw_arready_next = '0;

                case (state_counter)
                    0: desc_base_addr_next = m_axil.rdata;
                    1: desc_length_next = m_axil.rdata;
                    2:; // reserved
                    3: begin
                        desc_next_addr_next = m_axil.rdata;
                        state_next = READ;
                        state_counter_next = '0;
                    end
                endcase
            end
        end

        READ: begin
            m_axil.araddr = desc_base_addr + state_counter;

            m_axil.arvalid = saw_arready ? '0 : '1;
            saw_arready_next = m_axil.arready ? '1 : saw_arready;

            m_axil.rready = '1;

            if (m_axil.rvalid) begin
                saw_arready_next = '0;
                state_next = STREAM;
                buffer_next = m_axil.rdata;
                width_convert_count_next = '0;
            end
        end


        STREAM: begin
            m_axis.tdata = buffer[AXIS_DATA_WIDTH*width_convert_count +: AXIS_DATA_WIDTH];
            m_axis.tkeep = '1;
            m_axis.tvalid = '1;
            m_axis.tlast = '0;
            m_axis.tid = '0;
            m_axis.tdest = '0;
            m_axis.tuser = '0;

            if (m_axis.tready) begin
                width_convert_count_next = width_convert_count + 1;

                if (width_convert_count_next == AXIL_DATA_WIDTH/AXIS_DATA_WIDTH) begin
                    state_counter_next = state_counter + width_convert_count_next;
                    if (state_counter_next >= desc_length) begin
                        m_axis.tlast = '1;
                        state_next = CONTINUE;
                    end else begin
                        state_next = READ;
                    end
                end
            end
        end

        CONTINUE: begin
            desc_count_next = desc_count + 1;
            hwif_in.curr_addr.d.next = current_desc_addr;

            state_counter_next = '0;
            if (current_desc_addr == hwif_out.end_addr.d.value) begin
                state_next = IDLE;
            end else begin
                state_next = DESC_FETCH;
                current_desc_addr_next = desc_next_addr;
                hwif_in.curr_addr.d.next = current_desc_addr;
            end
        end

        default: $error("Bad State");

    endcase
end


endmodule